#!/usr/bin/env python3
"""
Main entrypoint for Uppsala pyub --

Query Uppsala University's holdings via cli or script.

The program runs a qeury, displays results to the user and performs an action on the selected result.
"""
import argparse
from uppsala_pyub.bibtex import make_bibtex
from uppsala_pyub.query import run_search
from uppsala_pyub.handle_results import display_results




def links(xml_data, output_location='.'):
    """
    place holder for eventual action... will probably be additional submodule.
    """
    raise NotImplementedError("`list_links` is not yet implemented.")

def citation(xml_data, output_location='.'):
    """
    place holder for eventual action... will probably be additional submodule.
    """
    raise NotImplementedError("`generate_citation` is not implemented. Fetch bibtex instead.")

def location(xml_data, output_location='.'):
    """
    place holder for eventual action... will probably be additional submodule.
    """
    raise NotImplementedError("`show_physical_location` is not yet implemented.")



#if __name__ == '__main__':
def cli():
    """
    Run command line interface
    """
    citation_formats = {
        "apa-6": "APA (6th edition)",
        "apa-7": "APA (7th edition)",
        "chicago": "Chicago/Turabian (16th edition)",
        "harvard": "Harvard",
        # add rest of styles
    }
    precisions = {
        "contains": "term is contained in the target material",
        "exact": "term is an exact match to target material",
        "begins_with": "target material begins with term",
    }
    resource_types = [
        "books",
        "articles",
        "book_chapter",
    ]

    parser = argparse.ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(help="Chose an action", dest="action")

    # actions:
    ## Fetch a bibtex entru
    fetch_bibtex_parser = subparsers.add_parser("fetch_bibtex", aliases=["bib"], help="Download bibtex entry for the selected resource.")

    ## list links
    list_links_parser = subparsers.add_parser("list_links", aliases=["link"], help="List links to / about the selected resource.")

    ## Generate a citation string
    generate_citation_parser = subparsers.add_parser("generate_citation", aliases=["cit"], help="Generate a citation for the selected resource.")

    generate_citation_parser.add_argument("--format", type=str, choices=list(citation_formats.keys()), default="harvard", help=f"Desited format of the citation string. {citation_formats}")

    ## show locations
    show_location_parser = subparsers.add_parser("show_physical_locations", aliases=["loc"], help="Show physical locations for the selected resource.")

    # shared args
    parser.add_argument("-o", "--output-destination",
                        type=str,
                        default="./",
                        help="Where to write output (if you want to write output). Q")
    parser.add_argument("-c", "--creator",
                        type=str,
                        default=None,
                        help="Creator / Author string. Q")
    parser.add_argument("--creator-precision",
                        type=str,
                        default="contains",
                        choices=list(precisions.keys()),
                        help=f"How do you want to apply the search? {precisions} Q")
    parser.add_argument("-t", "--title",
                        type=str,
                        default=None,
                        help="Title of work. Q")
    parser.add_argument("--title-precision",
                        type=str,
                        default="contains",
                        choices=list(precisions.keys()),
                        help=f"How do you want to apply the search? {precisions} Q")
    parser.add_argument("-s", "--creation_from",
                        type=str,
                        default=None,
                        help="Creation date range start -- YYYY or YYYY-mm-dd -- F")
    parser.add_argument("-e", "--creation_to",
                        type=str,
                        default=None,
                        help="Creation date range end -- YYYY or YYYY-mm-dd -- F")
    parser.add_argument("-T", "--resource-type",
                        type=str,
                        default=None,
                        choices=resource_types,
                        help=f"Select a specific type of resource to find. {resource_types} F")
    parser.add_argument("--join-by",
                        type=str,
                        default="AND",
                        choices=["AND", "OR"],
                        help=f"How to join the query parameters. [\"AND\", \"OR\"]")
    parser.add_argument("-q", "--query",
                        type=str,
                        default=None,
                        help="Enter a custom query")
    parser.add_argument("-l", "--language",
                        type=str,
                        default="en",
                        choices=["en","sv"],
                        help="Language of search interface. [\"en\", \"sv\"]")
    args = parser.parse_args()

    #main program
    programs = {
        "fetch_bibtex": make_bibtex,
        "bib": make_bibtex,
        "list_links": links,
        "generate_citation": citation,
        "show_physical_locations": location,
    }
    result_soup, url = run_search(**vars(args))
    selected = display_results(result_soup, url)
    if selected is not None:
        programs[args.action](selected, output_location=args.output_destination)




from lxml import etree
from uppsala_pyub.soup import make_soup
import re
import requests
print(etree.__file__)



def fetch_namespace():
    """
    return namespace dict for working with metadata xml
    """
    return {
            "xml":"{http://www.exlibrisgroup.com/xsd/primo/primo_nm_bib}",
            "sear": "{http://www.exlibrisgroup.com/xsd/jaguar/search}"
        }


def get_yatt(root, ns):
    """
    Returns year, author list, title, and resource type.

    Args:
        root: elementtree object -- parsed entry data
        ns: namespace dict
    """
    y, a, t, T = None, None, None, ''
    try:
        y = root.find(f"{ns['xml']}addata/{ns['xml']}date").text
    except:
        y = ""
    try:
        a = [_.text.strip() for _ in root.findall(f"{ns['xml']}addata/{ns['xml']}au") if _.text.strip() != '']
        assert a is not None
        assert len(a) != 0
    except:
        a = [_.text for _ in root.findall(f"{ns['xml']}addata/{ns['xml']}addau")]

    t = root.find(f"{ns['xml']}display/{ns['xml']}title").text
    try:
        T = root.find(f"{ns['xml']}display/{ns['xml']}rtype").text
    except:
        try:
            T = root.find(f"{ns['xml']}display/{ns['xml']}type").text
        except: pass
    return y, a, t, T


def display_results(soup, url):
    """
    Displays the results of a query to the user.

    Args:
        soup: bs4 object -- parsed xml result of query
        url (str): url with search parameters

    Returns:
        selection: etree object -- parsed metadata about selected entry
    """
    def _offset_url(url, value=10, operation="add"):
        def _add(mo, value=value):
            return f"offset={int(mo.group(1))+value}"

        def _subtract(mo, value=value):
            n = int(mo.group(1))-value
            if n < 0:
                n = 0
            return f"offset={n}"

        if operation == "add":
            url = re.sub(r'offset=(\d{1,5})', _add, url)
        elif operation == "subtract":
            url = re.sub(r'offset=(\d{1,5})', _subtract, url)
        else:
            raise ValueError(f"{operation} is not a valid operation. Choose add or subtract.")
        return url

    def _select(D):
        print("\nIs one of these what you're looking for? Select an entry by entering the value from the `Nr` column. or...")
        print("  - Enter `n` to see the next batch of results.")
        if "1" not in D:
            print("  - Enter `p` to see the previous batch of results.")
        print("  - Enter `a` to abort / skip this search attempt.")
        print("  - Enter `r` to reload this batch.")
        in_ = input("Your input: ")
        if in_ in D:
            return D[in_], None
        elif in_ in ["n", "p", "a", "r"]:
            return None, in_
        else:
            print("Invalid entry. Try again!")
            return _select(D)

    def _parse_datafile(data_url):
        data_url = data_url.replace("&amp;", "&")
        #print("~~", data_url)
        response = requests.get(data_url)
        root = etree.fromstring(response.content)
        return root, fetch_namespace()

    def _show(soup):
        items = soup.find_all("prm-brief-result-container")
        D = {}
        print("{:5}| {:12}| {:12}| {:60}| {:70}".format("Nr", "type", "year", "authors", "title"))
        for item in items:
            nr = item.find("span", class_="list-item-count").getText()
            root, ns = _parse_datafile(item.find("span", {"data-url": True}).attrs.get("data-url"))
            D[nr] = root
            year, authors, title, type_ = get_yatt(root, ns)
            print(f"{nr:5}| {type_[:10]:12}| {year[:10]:12}| {'; '.join(authors)[:58]:60}| {title[:68]:70}")
        return D


    selection = None
    action = None
    while selection == None:
        if action is not None:
            if action == "n":
                url = _offset_url(url)
                #print("new url:", url)
                soup = make_soup(url)
            elif action == "p":
                url = _offset_url(url, operation="subtract")
                #print("new url:", url)
                soup = make_soup(url)
            elif action == 'r':
                action = None
                soup = make_soup(url)
            elif action == "a":
                return None
        D = _show(soup)
        selection, action = _select(D)
        if selection is not None:
            return selection







from uppsala_pyub.handle_results import (
    fetch_namespace,
    get_yatt,
)
import re




def make_bibtex(xml_data, output_location='.'):
    ns = fetch_namespace()

    def _write_bibtex(bibtex, bibtex_key):
        with open(f"{output_location}/{bibtex_key}.bib", "w+") as out:
            bibtex = [f"{b},\n" if i < len(bibtex)-1 else f"{b}" for i, b in enumerate(bibtex)]
            [out.write(b) for b in bibtex]
            out.write("\n}")

    def _generate_key(author_list, year):
        lasts = [_.split(',')[0].replace(" ", "").strip() for _ in author_list]
        if len(lasts) == 1:
            key = lasts[0].lower() + year[:4]
        elif len(lasts) < 4:
            key = ''.join(lasts)+year[:4]
        else:
            key = lasts[0] + "EA" + year[:4]
        return key

    def _booktitle(bibtex):
        try:
            part_of = xml_data.find(f"{ns['xml']}display/{ns['xml']}ispartof").text
            assert part_of is not None
            assert part_of != ""
        except:
            return bibtex
        else:
            try:
                title, e, rest = re.split(r'\[(e|E)ds?\],', part_of)
            except:
                title = part_of
                e, rest = None, None
            for line in bibtex:
                if "    editor = {" in line:
                    editor = line.split("{")[1].strip("}")
                    for e in editor.split("AND"):
                        e = e.strip()
                        title.replace(e, "")
                    title.strip().strip("&").strip("and").strip().strip(",")
            bibtex.append(f"    booktitle = {{{title}}}")
            if rest is not None:
                pages = re.search(r'pp\s(\d{1,5}(-\d{1,5}))', rest)
                if pages is not None:
                    bibtex.append(f"   pages = {{{pages.group(1)}}}")
                plapub = re.search(r'(\S+:.*),', rest)
                if plapub is not None:
                    place, publisher = [_.strip() for _ in plapub.group(1).split(":")]
                    bibtex.append(f"    location = {{{place}}}")
                    bibtex.append(f"    publisher = {{{publisher}}}")
        return bibtex

    def _doi(bibtex):
        try:
            doi = xml_data.find(f"{ns['xml']}addata/{ns['xml']}doi").text
            assert doi is not None
            assert doi != ""
        except:
            return bibtex
        else:
            bibtex.append(f"    doi = {{{doi.strip()}}}")
        return bibtex

    def _editor(bibtex):
        try:
            editors = xml_data.find(f"{ns['xml']}display/{ns['xml']}contributor").text
            assert editors is not None
            assert editors is not None
        except:
            return bibtex
        else:
            editors = ' AND '.join([_.strip() for _ in editors.split(';')])
            bibtex.append(f"    editor = {{{editors.split()}}}")
        return bibtex

    def _isbn(bibtex):
        try:
            ISBN = xml_data.find(f"{ns['xml']}search/{ns['xml']}isbn").text
            assert ISBN is not None
            assert ISBN != ""
        except:
            try:
                identifier = xml_data.find(f"{ns['xml']}display/{ns['xml']}identifier").text
                m = re.search(r'\$\$CISBN\$V([\d-]+);', identifier)
                if m is not None:
                    ISBN = m.group(1)
                assert ISBN is not None
                assert ISBN != ""
            except:
                return bibtex
            else:
                bibtex.append(f"    isbn = {{{ISBN}}}")
        else:
            bibtex.append(f"    isbn = {{{ISBN}}}")
        return bibtex

    def _issue(bibtex):
        try:
            issue = xml_data.find(f"{ns['xml']}addata/{ns['xml']}issue").text
            assert issue is not None
            assert issue != ""
        except:
            return bibtex
        else:
            issue = re.sub(r'-{1}', "--", issue)
            bibtex.append(f"    issue = {{{issue.strip()}}}")
        return bibtex

    def _journal(bibtex):
        try:
            jtitle = xml_data.find(f"{ns['xml']}addata/{ns['xml']}jtitle").text
            assert jtitle is not None
            assert jtitle != ""
        except:
            return bibtex
        else:
            bibtex.append(f"    journal = {{{jtitle.strip()}}}")
        return bibtex

    def _pages(bibtex):
        try:
            pages = xml_data.find(f"{ns['xml']}addata/{ns['xml']}pages").text
            assert pages is not None
            assert pages != ""
        except:
            return bibtex
        else:
            pages = re.sub(r'-{1}', "--", pages)
            bibtex.append(f"    pages = {{{pages.strip()}}}")
        return bibtex

    def _place(bibtex):
        try:
            place = xml_data.find(f"{ns['xml']}display/{ns['xml']}place").text
            place = place.replace(":", "")
            place = place.strip()
            assert place is not None
            assert place != ''
        except:
            return bibtex
        else:
            bibtex.append(f"    location = {{{place}}}")
        return bibtex

    def _publisher(bibtex):
        for line in bibtex:
            if "    publisher = {" in line:
                return bibtex
        try:
            publisher = xml_data.find(f"{ns['xml']}display/{ns['xml']}publisher").text
            publisher = publisher.split(":")[-1]
            publisher = publisher.strip()
            assert publisher is not None
            assert publisher != ""
        except:
            return bibtex
        else:
            bibtex.append(f"    publisher = {{{publisher}}}")
        return bibtex

    def _volume(bibtex):
        try:
            volume = xml_data.find(f"{ns['xml']}addata/{ns['xml']}volume").text
            assert volume is not None
            assert volume != ""
        except:
            return bibtex
        else:
            bibtex.append(f"    volume = {{{volume.strip()}}}")
        return bibtex


    resource_types = {
        "books": "book",
        "book": "book",
        "article": "article",
        "book_chapter": "incollection",
    }

    fields = {
        "book": [_publisher, _place, _isbn],
        "article": [_journal, _volume, _issue, _pages, _doi],
        "incollection": [_editor, _booktitle, _publisher, _place, _pages, _isbn],
    }

    year, author_list, title, resource_type = get_yatt(xml_data, ns)
    bibtex = []
    if resource_type not in resource_types:
        raise ValueError(f"Unknown resource type {resource_type}. Select from {resource_types}.")

    rt = resource_types[resource_type]

    bibtex_key = _generate_key(author_list, year)
    bibtex.append(f"@{resource_types[resource_type]}" + "{" f"{bibtex_key}")
    bibtex.append(f"    title = {{{title.strip()}}}")
    bibtex.append(f"    year = {{{year[:4]}}}")
    bibtex.append(f"    author = {{{' AND '.join(author_list)}}}")
    for field in fields[rt]:
    #    print(field.__qualname__)
        bibtex = field(bibtex)
    #    print(bibtex)

    _write_bibtex(bibtex, bibtex_key)
    return bibtex_key





#!/usr/bin/env python3
"""
Query the Uppsala University Library search service
"""
from uppsala_pyub.soup import make_soup
import re




def query_template():
    """
    Return a dictionary of query key-value pairs and default values.
    """
    return {
        "output_destination": "./",
        "creator": None,
        "creator_precision": "contains",
        "title": None,
        "title_precision": "contains",
        "creation_from": None,
        "creation_to": None,
        "resource_type": None,
        "join_by": "AND",
        "query": None,
        "language": "en",
    }


def cap_query(query):
    """
    fill in all the missing variable to make a complete query dict
    """
    t = query_template()
    for k, v in t.items():
        if k not in query:
            query[k] = v
    return query


def format_date_string(datestring, startdate=True):
    """
    Format the datestring to YYYY-mm-dd format.

    Startdate provede 01-01 to a year, else 12-31

    This function assumes input is in YYYY format. Fully formed dates are returned as is, and in YYYY-mm, the month will be overwritten according to the startdate parameter.
    """
    pat = re.compile(r'\d{4}-\d{2}-\d{2}')
    if pat.search(datestring) is not None:
        return datestring
    try:
        _ = int(datestring[:4])
        if startdate:
            formatted_datestring = f"{datestring[:4]}-01-01"
        else:
            formatted_datestring = f"{datestring[:4]}-12-31"
        return formatted_datestring
    except:
        ValueError(f"Bad datestring: I don't know how to format your datestring entry {datestring}.")


def format_creator_names(names):
    """
    returns a list of creators, e.g. when passed a multi author citation `Smith and Doe`
    """
    formatted_names = []
    pat = re.compile(r'(,|and|&)')
    names = pat.sub("|", names)
    formatted_names.extend([_.strip() for _ in names.split("|") if _.strip() != ''])
    return formatted_names


def search_url(**kwargs):
    """
    Construct search URL based on kwarg parameters.
    """
    base_url = "https://uub.primo.exlibrisgroup.com/discovery/search?"
    tabscope = f"tab=Everything&search_scope=MyInst_and_CI&vid=46LIBRIS_UUB:UUB&lang={kwargs.get("language")}&mode=advanced&offset=0"
    filters = []

    if kwargs.get("query") is None:
        queries = []
        if kwargs.get("creator") is not None:
            creators = format_creator_names(kwargs.get("creator"))
            for c in creators:
                queries.append("query=creator," + kwargs.get('creator_precision') + "," + c)
        if kwargs.get("title") is not None:
            queries.append("query=title," + kwargs.get("title_precision") + "," + kwargs.get("title"))
        query = f",{kwargs.get('join_by')}&".join(queries)
    else:
        query = kwargs.get("query")
    if kwargs.get("resource_type") is not None:
        filters.append("pfilter=rtype,exact," + kwargs.get("resource_type"))
    if kwargs.get("creation_from") is not None:
        filters.append(f"pfilter=dr_s,exact,{format_date_string(kwargs.get('creation_from'))}")
        if kwargs.get("creation_to") is None:
            filters.append(f"pfilter=dr_e,exact,{format_date_string(kwargs.get('creation_from'), startdate=False)}")
        else:
            filters.append(f"pfilter=dr_e,exact,{format_date_string(kwargs.get('creation_to'), startdate=False)}")
    url = base_url + query
    if filters is not None and len(filters) > 0:
        url = url + ",AND&" + ",AND&".join(filters)
    url = url + ",AND&" + tabscope
    return url


def run_search(**kwargs):
    """
    Run a search based on key-word arguments. See ./cli.py for a list of accepted keywords.

    Returns:
        soup: (bs4 object) parsed result of the url
        url (str): UU UB search url
    """
    url = search_url(**kwargs)
    print("\n\n", url, "\n\n")
    soup = make_soup(url)
    return soup, url




from bs4 import BeautifulSoup as bs
from requests_html import HTMLSession




def make_soup(url):
    """
    Makes a bs4 soup object from the url

    Args:
        url:
    """
    s = HTMLSession()
    r = s.get(url)
    r.html.render()
    r = r.html
    return bs(r.__dict__.get('_html'), 'html.parser')
